# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import abc
import pyuavcan.transport


class Deduplicator(abc.ABC):
    @abc.abstractmethod
    def should_accept_transfer(
        self, iface_id: int, transfer_id_timeout: float, transfer: pyuavcan.transport.TransferFrom
    ) -> bool:
        """
        The iface-ID is an arbitrary integer that is unique within the redundant group identifying the transport
        instance the transfer was received from.
        It could be the index of the redundant interface (e.g., 0, 1, 2 for a triply-redundant transport),
        or it could be something else like a memory address of a related object.
        Embedded applications usually use indexes, whereas in PyUAVCAN it may be more convenient to use ``id()``.

        The transfer-ID timeout is specified in seconds. It is used to handle the case of a node restart.
        """
        raise NotImplementedError



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# noinspection PyUnresolvedReferences
"""
The application module contains application-level protocol entities.
This module is not imported automatically because:

- Some applications may choose to rely on the presentation-level API directly instead of using this higher-level module,
  in which case having it imported would be counter-productive due to increased initialization time, memory footprint,
  and possible reliability issues. The presentation level is the main abstraction layer provided by this library.
  The functionality of this application-level module is built on top of the public user-facing API of the
  presentation layer (no internal interfaces between modules exist). See :mod:`pyuavcan.presentation`.

- The module depends on DSDL generated packages, particularly on that of the standard root namespace ``uavcan``,
  so this module cannot be imported until the required code is generated.

>>> import pyuavcan
>>> pyuavcan.transport   # Works.
<module ...>
>>> pyuavcan.application
Traceback (most recent call last):
    ...
AttributeError: module 'pyuavcan' has no attribute 'application'
>>> import pyuavcan.application  # Will fail unless the DSDL package "uavcan" is generated and importable.

Classes contained here affect the state of the bus by publishing data, responding to service requests, or otherwise.
It is expected that some applications may need to complete early initialization procedures before
they are ready to begin interaction with the outside world.
Hence, many classes in this module are equipped with a method ``start()``,
which must be invoked once to bring the instance into a functional state.
The instance will remain operational until ``close()`` is invoked on it.
Both methods are idempotent.

The main entity of this module is the class :class:`pyuavcan.application.Node`.
There are several nested submodules,
each dedicated to a particular *application-level function* of the UAVCAN protocol,
such as :mod:`pyuavcan.application.heartbeat_publisher`;
each such module shall be imported explicitly.
Read the UAVCAN specification for background.
"""

from ._node import Node as Node, NodeInfo as NodeInfo



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This convenience module implements forwarding between the standard messages ``uavcan.diagnostic.Record``
over the standard subject-ID and the local logging facilities.
"""

import logging
import textwrap
from uavcan.diagnostic import Record_1_1 as Record
from uavcan.diagnostic import Severity_1_0 as Severity
import pyuavcan


_logger = logging.getLogger(__name__)


class DiagnosticSubscriber:
    """
    When started, subscribes to ``uavcan.diagnostic.Record``
    and forwards every received message into the standard Python ``logging`` facility.
    The logger name is that of the current module.

    The log level is mapped as follows:

    +-------------------------------+-------------------+
    | ``uavcan.diagnostic.Severity``| ``logging`` level |
    +===============================+===================+
    | TRACE                         | INFO              |
    +-------------------------------+-------------------+
    | DEBUG                         | INFO              |
    +-------------------------------+-------------------+
    | INFO                          | INFO              |
    +-------------------------------+-------------------+
    | NOTICE                        | INFO              |
    +-------------------------------+-------------------+
    | WARNING                       | WARNING           |
    +-------------------------------+-------------------+
    | ERROR                         | ERROR             |
    +-------------------------------+-------------------+
    | CRITICAL                      | CRITICAL          |
    +-------------------------------+-------------------+
    | ALERT                         | CRITICAL          |
    +-------------------------------+-------------------+

    Such logging behavior is especially convenient for various CLI tools and automation scripts where the user will not
    need to implement additional logic to see log messages from the network.
    """

    _LEVEL_MAP = {
        Severity.TRACE: logging.INFO,
        Severity.DEBUG: logging.INFO,
        Severity.INFO: logging.INFO,
        Severity.NOTICE: logging.INFO,
        Severity.WARNING: logging.WARNING,
        Severity.ERROR: logging.ERROR,
        Severity.CRITICAL: logging.CRITICAL,
        Severity.ALERT: logging.CRITICAL,
    }

    def __init__(self, presentation: pyuavcan.presentation.Presentation):
        self._sub_record = presentation.make_subscriber_with_fixed_subject_id(Record)

    def start(self) -> None:
        self._sub_record.receive_in_background(self._on_message)

    def close(self) -> None:
        self._sub_record.close()

    async def _on_message(self, msg: Record, meta: pyuavcan.transport.TransferFrom) -> None:
        node_id = meta.source_node_id if meta.source_node_id is not None else "anonymous"
        diag_text = msg.text.tobytes().decode("utf8", errors="replace")
        log_text = (
            f"uavcan.diagnostic.Record: node={node_id} severity={msg.severity.value} "
            + f"ts_sync={msg.timestamp.microsecond * 1e-6:0.6f} ts_local={meta.timestamp}:\n"
            + diag_text
        )
        level = self._LEVEL_MAP.get(msg.severity.value, logging.CRITICAL)
        _logger.log(level, log_text)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

r"""
Submodule import policy
+++++++++++++++++++++++

The following submodules are auto-imported when the root module ``pyuavcan`` is imported:

- :mod:`pyuavcan.dsdl`

- :mod:`pyuavcan.transport`, but not concrete transport implementation submodules.
  For example, if you need the CAN transport, import :mod:`pyuavcan.transport.can` manually.

- :mod:`pyuavcan.presentation`

- :mod:`pyuavcan.util`

The submodule :mod:`pyuavcan.application` is not auto-imported because in order to have it imported
the DSDL-generated package ``uavcan`` containing the standard data types must be generated first.

There are no internal (hidden) API between the submodules; they rely only on each other's public API.


Log level override
++++++++++++++++++

The environment variable ``PYUAVCAN_LOGLEVEL`` can be set to one of the following values to override
the library log level:

- ``CRITICAL``
- ``FATAL``
- ``ERROR``
- ``WARNING``
- ``INFO``
- ``DEBUG``

If not set, the log level is determined following the regular policies of the Python's standard ``logging`` library.
"""

import os as _os
import sys as _sys


with open(_os.path.join(_os.path.dirname(__file__), "VERSION")) as _version:
    __version__ = _version.read().strip()
__version_info__ = tuple(map(int, __version__.split(".")[:3]))
__author__ = "UAVCAN Consortium"
__copyright__ = "Copyright (c) 2019 UAVCAN Consortium"
__email__ = "consortium@uavcan.org"
__license__ = "MIT"


UAVCAN_SPECIFICATION_VERSION = 1, 0
"""
Version of the UAVCAN protocol implemented by this library, major and minor.
Use this value to populate the corresponding field in ``uavcan.node.GetInfo.Response``.
"""


if _sys.version_info[:2] < (3, 7):  # pragma: no cover
    raise RuntimeError("A newer version of Python is required")


# Configure logging if requested via environment variable.
# Accepted values: CRITICAL, ERROR, WARNING, INFO, DEBUG
_log_level_from_env = _os.environ.get("PYUAVCAN_LOGLEVEL")
if _log_level_from_env is not None:
    import logging as _logging

    _logging.basicConfig(
        format="%(asctime)s %(process)5d %(levelname)-8s %(name)s: %(message)s", level=_log_level_from_env
    )
    _logging.getLogger(__name__).setLevel(_log_level_from_env)
    _logging.getLogger(__name__).info("Log config from env var; level: %r", _log_level_from_env)


# The sub-packages are imported in the order of their interdependency.
import pyuavcan.util as util  # pylint: disable=wrong-import-position  # noqa
import pyuavcan.dsdl as dsdl  # pylint: disable=wrong-import-position  # noqa
import pyuavcan.transport as transport  # pylint: disable=wrong-import-position  # noqa
import pyuavcan.presentation as presentation  # pylint: disable=wrong-import-position  # noqa



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import logging
import uavcan.node
import pyuavcan
import pyuavcan.application
import pyuavcan.application.heartbeat_publisher
import pyuavcan.application.diagnostic


NodeInfo = uavcan.node.GetInfo_1_0.Response


_logger = logging.getLogger(__name__)


class Node:
    """
    This is the top-level abstraction representing a UAVCAN node on the bus.
    This class is just a minor addition on top of the lower-level abstractions of the library
    implementing commonly-used/mandatory functions of the protocol such as heartbeat reporting and responding
    to node info requests ``uavcan.node.GetInfo``.

    Start the instance when initialization is finished by invoking :meth:`start`.

    This class automatically instantiates the following application-level function implementations:

    - :class:`pyuavcan.application.heartbeat_publisher.HeartbeatPublisher` (see :attr:`heartbeat_publisher`).

    Additionally, if enabled via the corresponding constructor arguments, optional application-level function
    implementations are instantiated as described in the constructor documentation.
    """

    def __init__(
        self,
        presentation: pyuavcan.presentation.Presentation,
        info: NodeInfo,
        with_diagnostic_subscriber: bool = False,
    ):
        """
        :param presentation: The node takes ownership of the supplied presentation controller.
            Ownership here means that the controller will be closed (along with all sessions and other resources)
            when the node is closed.

        :param info: The info structure is sent as a response to requests of type ``uavcan.node.GetInfo``;
            the corresponding server instance is established and run by the node class automatically.

        :param with_diagnostic_subscriber: If True, an instance of
            :class:`pyuavcan.application.diagnostic.DiagnosticSubscriber` will be constructed to channel
            standard UAVCAN diagnostic messages into the local Python logging facility.
        """
        self._presentation = presentation
        self._info = info
        self._heartbeat_publisher = pyuavcan.application.heartbeat_publisher.HeartbeatPublisher(self._presentation)
        self._srv_info = self._presentation.get_server_with_fixed_service_id(uavcan.node.GetInfo_1_0)

        self._diagnostic_subscriber = (
            pyuavcan.application.diagnostic.DiagnosticSubscriber(self._presentation)
            if with_diagnostic_subscriber
            else None
        )

        self._started = False

    @property
    def presentation(self) -> pyuavcan.presentation.Presentation:
        """Provides access to the underlying instance of :class:`pyuavcan.presentation.Presentation`."""
        return self._presentation

    @property
    def info(self) -> NodeInfo:
        """Provides access to the local node info structure. See :class:`pyuavcan.application.NodeInfo`."""
        return self._info

    @property
    def heartbeat_publisher(self) -> pyuavcan.application.heartbeat_publisher.HeartbeatPublisher:
        """Provides access to the heartbeat publisher instance of this node."""
        return self._heartbeat_publisher

    def start(self) -> None:
        """
        Starts the GetInfo server in the background, the heartbeat publisher, etc.
        Those will be automatically terminated when the node is closed.
        Does nothing if already started.
        """
        if not self._started:
            self._srv_info.serve_in_background(self._handle_get_info_request)
            self._heartbeat_publisher.start()
            if self._diagnostic_subscriber is not None:
                self._diagnostic_subscriber.start()
            self._started = True

    def close(self) -> None:
        """
        Closes the underlying presentation instance, application-level functions, and all other entities.
        Does nothing if already closed.
        """
        try:
            self._heartbeat_publisher.close()
            self._srv_info.close()
            if self._diagnostic_subscriber is not None:
                self._diagnostic_subscriber.close()
        finally:
            self._presentation.close()

    async def _handle_get_info_request(
        self, _: uavcan.node.GetInfo_1_0.Request, metadata: pyuavcan.presentation.ServiceRequestMetadata
    ) -> NodeInfo:
        _logger.debug("%s got a node info request: %s", self, metadata)
        return self._info

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, info=self._info, heartbeat=self._heartbeat_publisher.make_message(), presentation=self._presentation
        )
